from typing_extensions import Any, Literal, Sequence, AsyncIterable, Union
from abc import abstractmethod
from dataclasses import dataclass, field
from datetime import datetime
from decimal import Decimal

from trading_sdk.core import SDK

@dataclass(kw_only=True, frozen=True)
class Flow:
  id: str | None = None
  asset: str
  change: Decimal
  price: Decimal | None = None
  time: datetime
  event_id: str | None = None
  event_tag: str | None = None
  raw: Any
  source: str

@dataclass(kw_only=True, frozen=True)
class BaseEvent:
  tag: str
  id: str | None = None
  time: datetime
  raw: Any
  source: str

  def flow(self, asset: str, change: Decimal) -> Flow:
    return Flow(
      time=self.time,
      asset=asset,
      change=change,
      raw=self,
      event_id=self.id,
      event_tag=self.tag,
      source=f'generated:{self.tag}',
    )

  @property
  def flows(self) -> Sequence[Flow]:
    return []

@dataclass(kw_only=True, frozen=True)
class SpotTrade(BaseEvent):
  tag: Literal['spot_trade'] = 'spot_trade'
  qty: Decimal
  price: Decimal
  base: str
  quote: str
  liquidity: Literal['maker', 'taker'] | None = None
  side: Literal['buy', 'sell']
  fee: Decimal | None = None
  fee_asset: str | None = None

  @property
  def flows(self) -> Sequence[Flow]:
    sign = 1 if self.side == 'buy' else -1
    flows: list[Flow] = [
      self.flow(self.base, sign * self.qty),
      self.flow(self.quote, -sign * self.qty * self.price),
    ]
    if self.fee and self.fee_asset:
      flows.append(self.flow(self.fee_asset, -self.fee))
    return flows

@dataclass(kw_only=True, frozen=True)
class FuturesTrade(BaseEvent):
  tag: Literal['futures_trade'] = 'futures_trade'
  qty: Decimal
  price: Decimal
  instrument: str
  liquidity: Literal['maker', 'taker'] | None = None
  side: Literal['buy', 'sell']
  fee: Decimal | None = None
  fee_asset: str | None = None

  @property
  def flows(self) -> Sequence[Flow]:
    sign = 1 if self.side == 'buy' else -1
    flows: list[Flow] = [
      self.flow(self.instrument, sign * self.qty),
    ]
    if self.fee and self.fee_asset:
      flows.append(self.flow(self.fee_asset, -self.fee))
    return flows

@dataclass(kw_only=True, frozen=True)
class CryptoDeposit(BaseEvent):
  tag: Literal['crypto_deposit'] = 'crypto_deposit'
  qty: Decimal
  asset: str
  network: str
  tx_hash: str
  idx: int | None = None
  memo: str | None = None

  @property
  def flows(self) -> Sequence[Flow]:
    return [self.flow(self.asset, self.qty)]

@dataclass(kw_only=True, frozen=True)
class CryptoWithdrawal(BaseEvent):
  tag: Literal['crypto_withdrawal'] = 'crypto_withdrawal'
  qty: Decimal
  asset: str
  network: str
  tx_hash: str
  idx: int | None = None
  fee: Decimal | None = None
  fee_asset: str | None = None
  dst_address: str
  dst_memo: str | None = None

  @property
  def flows(self) -> Sequence[Flow]:
    flows: list[Flow] = [self.flow(self.asset, -self.qty)]
    if self.fee and self.fee_asset:
      flows.append(self.flow(self.fee_asset, -self.fee))
    return flows

@dataclass(kw_only=True, frozen=True)
class FiatDeposit(BaseEvent):
  tag: Literal['fiat_deposit'] = 'fiat_deposit'
  asset: str
  qty: Decimal
  """Amount of the asset credited."""
  fiat_currency: str
  fiat_amount: Decimal
  """Amount of the fiat currency deposited."""
  method: str

  @property
  def flows(self) -> Sequence[Flow]:
    return [self.flow(self.asset, self.qty)]

@dataclass(kw_only=True, frozen=True)
class FiatWithdrawal(BaseEvent):
  tag: Literal['fiat_withdrawal'] = 'fiat_withdrawal'
  asset: str
  qty: Decimal
  """Amount of the asset debited."""
  fiat_currency: str
  fiat_amount: Decimal
  """Amount of the fiat currency withdrawn."""
  method: str

  @property
  def flows(self) -> Sequence[Flow]:
    return [self.flow(self.asset, -self.qty)]

@dataclass(kw_only=True, frozen=True)
class Yield(BaseEvent):
  tag: Literal['yield'] = 'yield'
  asset: str
  qty: Decimal

  @property
  def flows(self) -> Sequence[Flow]:
    return [self.flow(self.asset, self.qty)]

@dataclass(kw_only=True, frozen=True)
class EthereumTx(BaseEvent):
  @dataclass(kw_only=True, frozen=True)
  class Execution:
    method_name: str | None = None
    """Function called (if any, if identified from source code)"""
    method_id: str
    """Function ID (if any)"""
    input: str
    """Input data (if any)"""

  tag: Literal['ethereum_transaction'] = 'ethereum_transaction'
  hash: str
  value: Decimal | None = None
  """Value received (if positive) or sent (if negative) [ETH]"""
  fee: Decimal | None = None
  """Fee paid [ETH]"""
  execution: Execution | None = None
  """Contract execution details (if any)"""

  @property
  def flows(self) -> Sequence[Flow]:
    flows: list[Flow] = []
    if self.value:
      flows.append(self.flow('ETH', self.value))
    if self.fee:
      flows.append(self.flow('ETH', -self.fee))
    return flows

Event = Union[
  SpotTrade, FuturesTrade, Yield,
  CryptoDeposit, CryptoWithdrawal,
  FiatDeposit, FiatWithdrawal,
  EthereumTx,
]

class History(SDK):
  @dataclass(kw_only=True, frozen=True)
  class History:
    flows: Sequence[Flow] = field(default_factory=list)
    events: Sequence[Event] = field(default_factory=list)

  @SDK.method
  @abstractmethod
  def history(self, start: datetime, end: datetime) -> AsyncIterable[History]:
    """Fetch your reporting history."""

  @SDK.method
  async def history_sync(self, start: datetime, end: datetime) -> History:
    """Fetch your reporting history without streaming."""
    flows: list[Flow] = []
    events: list[Event] = []
    async for page in self.history(start, end):
      flows.extend(page.flows)
      events.extend(page.events)
    return History.History(flows=flows, events=events)


from dataclasses import dataclass, field
from typing_extensions import Literal

from mexc_sdk import MEXC
from dydx_sdk import DYDX
from hyperliquid_sdk import Hyperliquid

from trading_sdk.core import UserError

from . import Market

MarketType = Literal['spot', 'perp']

@dataclass
class Markets:
  """Market loader with manual, in-memory caching."""

  _mexc: MEXC | None = field(default=None, init=False, repr=False)
  _dydx: DYDX | None = field(default=None, init=False, repr=False)
  _hyperliquid: Hyperliquid | None = field(default=None, init=False, repr=False)
  _hyperliquid_spot: Hyperliquid.Spot | None = field(default=None, init=False, repr=False)
  _hyperliquid_perp: dict[str | None, Hyperliquid.Perp] = field(default_factory=dict, init=False, repr=False)

  _spot_cache: dict[tuple[str, str], Market] = field(default_factory=dict, init=False, repr=False)
  _perp_cache: dict[tuple[str, str], Market] = field(default_factory=dict, init=False, repr=False)

  async def market(self, spec: str) -> Market:
    """Load a market from a specification string.

    Format: `<venue>:<spot|perp>:<instrument_id>`

    Examples:
    - `mexc:spot:BTCUSDT`
    - `mexc:perp:BTC_USDT`
    - `dydx:perp:BTC-USD`
    - `hyperliquid:perp:BTC`
    - `hyperliquid:perp:xyz:BTC`
    """
    venue, market_type, instrument_id = spec.split(':', 2)
    if market_type == 'spot':
      return await self.spot(venue, instrument_id)
    elif market_type == 'perp':
      return await self.perpetual(venue, instrument_id)
    else:
      raise UserError(f'Unsupported market type: {market_type!r}')

  async def spot(self, platfrom: str, instrument: str) -> Market:
    """Load a spot market.

    Formats:
    - mexc: ``BTCUSDT``
    - hyperliquid: ``BASE/QUOTE`` or ``BASE/QUOTE:INDEX`` (e.g. ``UBTC/USDC:142``)
    - dydx: not supported (raises)
    """
    cache_key = (platfrom, instrument)
    cached = self._spot_cache.get(cache_key)
    if cached is not None:
      return cached

    if platfrom == 'mexc':
      mexc = self._mexc_sdk()
      market = await mexc.spot(instrument)
    elif platfrom == 'hyperliquid':
      base, quote, index = self._parse_hyperliquid_spot_instrument(instrument)
      market = await self._hyperliquid_spot_market(base, quote, index)
    elif platfrom == 'dydx':
      raise UserError('dYdX does not support spot markets.')
    else:
      raise UserError(f'Unsupported platform for spot: {platfrom!r}')

    self._spot_cache[cache_key] = market
    return market

  async def perpetual(self, platfrom: str, instrument: str) -> Market:
    """Load a perpetual market.

    Formats:
    - mexc: ``BTC_USDT``
    - dydx: ``SUBACCOUNT:MARKET`` or ``MARKET`` (e.g. ``128:BTC-USD``)
    - hyperliquid: ``DEX:BASE`` or ``BASE`` (e.g. ``xyz:BTC`` or ``BTC``)
    """
    cache_key = (platfrom, instrument)
    cached = self._perp_cache.get(cache_key)
    if cached is not None:
      return cached

    if platfrom == 'mexc':
      mexc = self._mexc_sdk()
      market = await mexc.perp(instrument)
    elif platfrom == 'dydx':
      subaccount, market_id = self._parse_dydx_perp_instrument(instrument)
      market = await self._dydx_market(market_id, subaccount=subaccount)
    elif platfrom == 'hyperliquid':
      dex, base = self._parse_hyperliquid_perp_instrument(instrument)
      market = await self._hyperliquid_perp_market(base=base, dex=dex)
    else:
      raise UserError(f'Unsupported platform for perpetual: {platfrom!r}')

    self._perp_cache[cache_key] = market
    return market

  def _mexc_sdk(self) -> MEXC:
    if self._mexc is None:
      self._mexc = MEXC.new(None, None)
    return self._mexc

  async def _dydx_sdk(self) -> DYDX:
    if self._dydx is None:
      self._dydx = await DYDX.connect(mnemonic=None)
    return self._dydx

  async def _dydx_market(self, market: str, *, subaccount: int = 0) -> Market:
    dydx_client = await self._dydx_sdk()
    return await dydx_client.market(market, subaccount=subaccount)

  def _hyperliquid_sdk(self) -> Hyperliquid:
    if self._hyperliquid is None:
      self._hyperliquid = Hyperliquid.new(
        address=None,
        wallet=None,
      )
    return self._hyperliquid

  async def _hyperliquid_spot_client(self):
    if self._hyperliquid_spot is None:
      self._hyperliquid_spot = await self._hyperliquid_sdk().spot()
    return self._hyperliquid_spot

  async def _hyperliquid_perp_client(self, dex: str | None):
    if dex not in self._hyperliquid_perp:
      self._hyperliquid_perp[dex] = await self._hyperliquid_sdk().perp(dex)
    return self._hyperliquid_perp[dex]

  async def _hyperliquid_spot_market(
    self,
    base: str,
    quote: str,
    index: int | None,
  ) -> Market:
    spot = await self._hyperliquid_spot_client()
    market = spot.find(base, quote)
    if index is not None and market.asset_idx != index:
      raise ValueError(
        f"Expected market {base}/{quote} at index {index}, got {market.base_name}/{market.quote_name}",
      )
    return market

  async def _hyperliquid_perp_market(self, *, base: str, dex: str | None) -> Market:
    perp = await self._hyperliquid_perp_client(dex)
    if dex:
      base = f'{dex}:{base}'
    return perp.find(base)

  @staticmethod
  def _parse_hyperliquid_spot_instrument(instrument: str) -> tuple[str, str, int | None]:
    if ':' in instrument:
      ticker, index_str = instrument.split(':', 1)
      index = int(index_str)
    else:
      ticker, index = instrument, None
    base, quote = ticker.split('/', 1)
    return base, quote, index

  @staticmethod
  def _parse_hyperliquid_perp_instrument(instrument: str) -> tuple[str | None, str]:
    if ':' in instrument:
      dex, base = instrument.split(':', 1)
      dex = dex or None
      return dex, base
    return None, instrument

  @staticmethod
  def _parse_dydx_perp_instrument(instrument: str) -> tuple[int, str]:
    if ':' in instrument:
      subaccount_str, market = instrument.split(':', 1)
      return int(subaccount_str), market
    return 0, instrument



from decimal import Decimal, ROUND_HALF_DOWN, ROUND_FLOOR, ROUND_CEILING

Num = Decimal | str | float | int

def fmt_num(num: Num) -> str:
  out = f'{Decimal(num).normalize():f}'
  if out == '-0':
    return '0'
  else:
    return out

def round2tick(x: Decimal, tick_size: Decimal) -> Decimal:
  r = (x / tick_size).quantize(Decimal('1.'), rounding=ROUND_HALF_DOWN) * tick_size
  return r.normalize()

def trunc2tick(x: Decimal, tick_size: Decimal) -> Decimal:
  r = (x / tick_size).to_integral_value(rounding=ROUND_FLOOR) * tick_size
  return r.normalize()

def ceil2tick(x: Decimal, tick_size: Decimal) -> Decimal:
  r = (x / tick_size).to_integral_value(rounding=ROUND_CEILING) * tick_size
  return r.normalize()


from typing_extensions import Mapping, Any, Collection, Literal
from dataclasses import dataclass
from datetime import datetime
import re
import pandas as pd

Schema = Mapping[str|re.Pattern, type|re.Pattern]

def ensure_datetime(x, *, format: Literal['iso'] = 'iso') -> datetime:
  if isinstance(x, datetime):
    return x
  else:
    return datetime.fromisoformat(str(x))

@dataclass
class MissingColumn:
  column: str | re.Pattern
  expected_type: type | re.Pattern

  def __str__(self):
    return f'Column "{self.column}" is missing (expected type: {self.expected_type})'

@dataclass
class InvalidColumnType:
  column: str
  expected_type: type
  actual_type: type
  example_value: Any

  def __str__(self):
    return f'Column "{self.column}" has type {self.actual_type} (expected {self.expected_type})'

@dataclass
class InvalidColumnString:
  column: str
  expected_type: re.Pattern
  wrong_values: Collection[str]

  def __str__(self):
    return f'Column "{self.column}" has invalid values (expected it to match {self.expected_type.pattern}). Wrong values: {", ".join(list(self.wrong_values)[:10])}'

SchemaError = MissingColumn | InvalidColumnType | InvalidColumnString

class InvalidSchema(Exception):
  def __init__(self, errors: list[SchemaError], *args, **kwargs):
    super().__init__(*args, **kwargs)
    self.errors = errors

  def __str__(self):
    out = 'Schema Validation Error:\n'
    for error in self.errors:
      out += f'> {error}\n'
    return out

def find_key(df: pd.DataFrame, key: str | re.Pattern) -> str | None:
  keys = list(df.dtypes.keys())
  if isinstance(key, str):
    if key in keys:
      return key
  else:
    for k in keys:
      if key.match(k):
        return k

def validate_schema(df: pd.DataFrame, schema: Schema):
  errors: list[SchemaError] = []
  for expected_key, expected_type in schema.items():
    key = find_key(df, expected_key)
    if key is None:
      errors.append(MissingColumn(expected_key, expected_type))
      continue
    
    if isinstance(expected_type, re.Pattern):
      ok = df[key].astype(str).str.match(expected_type).all()
      if not ok:
        wrong_values = set[str](df[key][~df[key].astype(str).str.match(expected_type)].unique()) # type: ignore
        errors.append(InvalidColumnString(key, expected_type, wrong_values))
    
    elif expected_type is not Any and len(df) > 0:
      row = df.iloc[0]
      if not isinstance(row[key], expected_type):
        errors.append(InvalidColumnType(key, expected_type, type(row[key]), row[key]))

  if len(errors) > 0:
    raise InvalidSchema(errors)


from .exc import Error, NetworkError, ValidationError, UserError, AuthError, ApiError, LogicError
from .sdk import SDK, instrument, exponential_retry, log

__all__ = [
  'Error', 'NetworkError', 'ValidationError', 'UserError', 'AuthError', 'ApiError', 'LogicError',
  'SDK', 'instrument', 'exponential_retry', 'log',
]


from .numbers import Num, fmt_num, round2tick, trunc2tick, ceil2tick
from .csv import validate_schema, ensure_datetime, Schema
from . import csv

__all__ = [
  'Num', 'fmt_num',
  'round2tick', 'trunc2tick', 'ceil2tick',
  'validate_schema', 'ensure_datetime', 'Schema', 'csv',
]